#include <windows.h>
#include <math.h>
#include "buildbsp.h"
#include "map.h"
#include "sysutils.h"
#include "dialog.h"
#include "sysmain.h"
#include "resource.h"

int cuts;

/*
==================
=
= DivlineFromWorldline
=
==================
*/

void DivlineFromWorldline(divline_t *d, line_t *w)
{
    d->pt = w->p1;
    d->dx = w->p2.x - w->p1.x;
    d->dy = w->p2.y - w->p1.y;
}

/*
==================
=
= PointOnSide
=
= Returns side 0 (front), 1 (back), or -1 (colinear)
==================
*/

int PointOnSide(point_t* p, divline_t* l)
{
    double  dx, dy;
    double  px, py;
    double  len;
    double  val;

    len = sqrt(l->dx*l->dx + l->dy*l->dy);
    dx = l->dx / len;
    dy = l->dy / len;

    px = p->x - l->pt.x;
    py = p->y - l->pt.y;

    val = dx*py - px*dy; // 2D cross product

    if (val < -0.1)
        return 0;
    else if (val > 0.1)
        return 1;
    return -1;
}

/*
=============
=
= sign
=
= Returns -1, 0, or 1, based on the input sign
=
==============
*/

int sign(float i) // ĝLineOnSide()ł̏tf̎̂
{
    if (i < -0.1) // l00.1ɏCBConvexity Test G[̌mF(2012/02/26)
        return -1;
    else if (i > 0.1)
        return 1;
    return 0;
}

Storage<point_t>* colipoints = NULL;

/*
==================
=
= LineOnSide
=
= Returns side 0 / 1, or -2 if line must be split
= If the line is colinear, it will be placed on the front side if
= it is going the same direction as the dividing line
==================
*/

int/*boolean*/ LineOnSide(line_t* wl, divline_t* bl)
{
    int	    s1,s2;
    float   dx, dy;

    s1 = PointOnSide(&wl->p1, bl);
    s2 = PointOnSide(&wl->p2, bl);

    if (s1 == s2)
    {
        if (s1 == -1)
        {
            if (colipoints)
            {
                colipoints->addElement(&wl->p1);
                colipoints->addElement(&wl->p2);
            }
            // colinear, so see if the directions are the same
            dx = wl->p2.x - wl->p1.x;
            dy = wl->p2.y - wl->p1.y;
            if (sign(dx) == sign(bl->dx) && sign(dy) == sign(bl->dy))
                return 0;
            return 1;
        }
        return s1;
    }
    if (s1 == -1)
    {
        if (colipoints)
            colipoints->addElement(&wl->p1);
        return s2;
    }
    if (s2 == -1)
    {
        if (colipoints)
            colipoints->addElement(&wl->p2);
        return s1;
    }

    return -2;
}

/*
===============
=
= InterceptVector
=
= Returns the fractional intercept point along first vector
===============
*/

float InterceptVector(divline_t* v1, divline_t* v2) // Real-Time Rendering SE p616 Q
{
    float   frac, num, den;

    den = v1->dx*(-v2->dy) + v1->dy*v2->dx;
    if (den == 0)
        Error("InterceptVector: parallel");
    num = (v2->pt.x - v1->pt.x)*(-v2->dy) + (v2->pt.y - v1->pt.y)*v2->dx;
    frac = num / den;

    if (frac <= 0.0 || frac >= 1.0)
        Debug("InterceptVector: intersection outside line\n");
		
    return frac;
}

/*
==================
=
= CutLine
=
= Truncates the given worldline to the front side of the divline
= and returns the cut off back side in a newly allocated worldline
==================
*/

line_t* CutLine(line_t* wl, divline_t* bl)
{
    int         side;
    line_t*     new_p;
    divline_t   wld;
    float       frac;
    point_t     intr;
    float       offset;

    cuts++;
    DivlineFromWorldline(&wld, wl);
    new_p = (line_t *)malloc(sizeof(line_t));
    memset(new_p,0,sizeof(*new_p)); // sKv?
    *new_p = *wl; 

    frac = InterceptVector(&wld, bl);
    intr.x = wld.pt.x + wld.dx*frac;
    intr.y = wld.pt.y + wld.dy*frac;
    offset = wl->offset + frac*(float)sqrt(wld.dx*wld.dx+wld.dy*wld.dy);
    side = PointOnSide(&wl->p1, bl);
    if (side == 0)
    {   // line starts on front side
        wl->p2 = intr;
        new_p->p1 = intr;
        new_p->offset = offset;
    }
    else
    {   // line starts on back side
        wl->p1 = intr;
        wl->offset = offset;
        new_p->p2 = intr;
    }

    colipoints->addElement(&intr);

    return new_p;
}

/*
================
=
= EvaluateSplit
=
= Returns a number grading the quality of a split along the givent line
= for the current list of lines.  Evaluation is halted as soon as it is
= determined that a better split already exists
= 
= A split is good if it divides the lines evenly without cutting many lines
= A horizontal or vertical split is better than a sloping split
=
= The LOWER the returned value, the better.  If the split line does not divide
= any of the lines at all, MAXINT will be returned
================
*/

int EvaluateSplit(Storage<line_t>* lines, line_t* spliton, int bestgrade)
{
    int             i,c,side;
    line_t*         line_p;
    divline_t       divline;
    int             frontcount, backcount, max, inc;
    int             grade;

    DivlineFromWorldline(&divline, spliton);

    frontcount = backcount = 0;
    c = lines->count();
    grade = 0;
	
    for (i=0 ; i<c ; i++)
    {
        line_p = lines->elementAt(i);
        if (line_p->side == -1)
            continue;
        if (line_p == spliton)
            side = 0;
        else
            side = LineOnSide(line_p, &divline);
        switch (side)
        {
        case 0:
            frontcount++;
            break;
        case 1:
            backcount++;
            break;
        case -2:
            frontcount++;
            backcount++;
            break;
        }

        max = MAX(frontcount,backcount);
        inc = (frontcount+backcount) - c;
        grade = max+inc*8;
        if (grade > bestgrade)
            return grade;       // might as well stop now
    }

    if (frontcount == 0 || backcount == 0)
        return _MAXINT;         // line does not partition at all

    return grade;
}

BOOL ClosePoints(point_t* p1, point_t* p2)
{
    float   dx, dy;

    dx = p2->x - p1->x;
    dy = p2->y - p1->y;

    if ((dx > -0.1f && dx < 0.1f) && (dy > -0.1f && dy < 0.1f))
        return TRUE;

    return FALSE;
}

void AddInvisibleLines(Storage<line_t>* lines, Storage<line_t>* frontlist, Storage<line_t>* backlist)
{
    int         i, j, c, numlines;
    point_t     *p1, *p2;
    line_t*     line_p;
    vertex_t    midpt;
    line_t      line;

    c = colipoints->count();
    numlines = lines->count();

    p1 = colipoints->elementAt(0);
    for (i=1 ; i<c ; i++)
    {
        p2 = colipoints->elementAt(i);

        if (ClosePoints(p1, p2))
        {
            p1 = p2;
            continue;
        }

        for (j=0 ; j<numlines ; j++)
        {
            line_p = lines->elementAt(j);
            if ((ClosePoints(&line_p->p1, p1) && ClosePoints(&line_p->p2, p2)) ||
                (ClosePoints(&line_p->p2, p1) && ClosePoints(&line_p->p1, p2)))
            {
                if (line_p->side == -1)
                    Debug("AddInvisibleLines: an invisible line coincide with the divline\n");
                break;
            }
        }
        if (j != numlines)
        {
            p1 = p2;
            continue;
        }

        midpt.x = p1->x + 0.5f*(p2->x - p1->x);
        midpt.y = p1->y + 0.5f*(p2->y - p1->y);
        if (!PointInSector(midpt))
        {
            p1 = p2;
            continue;
        }

        line.side = -1;
        line.linedef = -1;
        line.offset = 0;
        line.grouped = false;
        line.p1 = *p1;
        line.p2 = *p2;
        frontlist->addElement(&line);
        line.p1 = *p2;
        line.p2 = *p1;
        backlist->addElement(&line);

        p1 = p2;
    }
}

divline_t g_divline;

int pointcmp(const void* elem1, const void* elem2)
{
    point_t* p1 = (point_t *)elem1;
    point_t* p2 = (point_t *)elem2;
    float   dp1, dp2;

    dp1 = g_divline.dx*(p1->x - g_divline.pt.x) + g_divline.dy*(p1->y - g_divline.pt.y);
    dp2 = g_divline.dx*(p2->x - g_divline.pt.x) + g_divline.dy*(p2->y - g_divline.pt.y);

    if (dp1 - dp2 > 0)
        return 1;
    if (dp2 - dp1 > 0)
        return -1;
    return 0;
}

/*
================
=
= ExecuteSplit
=
= Actually splits the line list as EvaluateLines predicted
================
*/

void ExecuteSplit(Storage<line_t>* lines, line_t* spliton, Storage<line_t>* frontlist, Storage<line_t>* backlist)
{
    int         i,c,side;
    line_t      *line_p, *newline_p;
    divline_t   divline;

    DivlineFromWorldline(&divline, spliton);

    c = lines->count();

    colipoints = new Storage<point_t>;
    colipoints->addElement(&spliton->p1);
    colipoints->addElement(&spliton->p2);

    for (i=0 ; i<c ; i++)
    {
        line_p = lines->elementAt(i);
        if (line_p == spliton)
            side = 0;
        else
            side = LineOnSide(line_p, &divline);
        switch (side)
        {
        case 0:
            frontlist->addElement(line_p);
            break;
        case 1:
            backlist->addElement(line_p);
            break;
        case -2:
            newline_p = CutLine(line_p, &divline);
            frontlist->addElement(line_p);
            backlist->addElement(newline_p);
            free(newline_p);
            break;
        //default:
            //Error("ExecuteSplit: bad side");
        }
    }

    // r邾ȂPʃxNg̕Kv͂Ȃ
    // lƔf鎞̋el܂Ă鎞͕
    g_divline = divline;
    c = colipoints->count();
    qsort(colipoints->elementAt(0), c, sizeof(point_t), pointcmp);

    AddInvisibleLines(lines, frontlist, backlist);

    delete colipoints;
    colipoints = NULL;
}

/*
================
=
= BSPList
=
= Takes a storage of lines and recursively partitions the list
= Returns a bspnode_t
================
*/

bspnode_t* BSPList(Storage<line_t>* lines)
{
    bspnode_t*      node_p;
    int             i, c, step;
    int             v, bestv;
    line_t          *line_p, *bestline_p;
    Storage<line_t> *frontlist, *backlist;

    node_p = (bspnode_t *)malloc(sizeof(*node_p));
    memset(node_p, 0, sizeof(*node_p));

//
// find the best line to partition on 
//
    c = lines->count();
    bestv = _MAXINT;	
    bestline_p = NULL;
    step = (c/40)+1;    // set this to 1 for an exhaustive search
research:
    for (i=0 ; i<c ; i+=step)
    {
        line_p = lines->elementAt(i);
        if (line_p->side == -1) // invisible line ͕͌Ȃ̂Ŏc line ɂ̂͂킩Ă̂ŒׂȂĂH
            continue;
        v = EvaluateSplit(lines, line_p, bestv);
        if (v<bestv)
        {
            bestv = v;
            bestline_p = line_p;
        }
    }

//
// if none of the lines should be split, the remaining lines
// are convex, and form a terminal node
//
    if (bestv == _MAXINT)
    {
        if (step > 1)
        {
            step = 1;
            goto research;
        }
        node_p->lines = lines;
        return node_p;
    }

//
// divide the line list into two nodes along the best split line
//
    DivlineFromWorldline(&node_p->divline, bestline_p);

    frontlist = new Storage<line_t>;
    backlist = new Storage<line_t>;

    ExecuteSplit(lines, bestline_p, frontlist, backlist);

    // node builder OvOƂčꍇ́A܂胁[NSzKv͖
    delete lines;

//
// recursively divide the lists
//
    node_p->side[0] = BSPList(frontlist);
    node_p->side[1] = BSPList(backlist);

    return node_p;
}

/*
=====================
=
= MakeSegs
=
=====================
*/

Storage<line_t>*    segstore;

void MakeSegs(void)
{
    int         i;
    linedef_t*  line;
    point_t     pt;
    line_t      li;

    segstore = new Storage<line_t>;

    for (i=0, line = firstlinedef.next ; line ; line = line->next, i++)
    {
        pt.x = line->v1->x;
        pt.y = line->v1->y;
        li.p1 = pt;
        pt.x = line->v2->x;
        pt.y = line->v2->y;
        li.p2 = pt;
        li.linedef = i;
        li.side = 0;
        li.offset = 0;
        li.grouped = false;
        segstore->addElement(&li);

        if (line->sides[1].sec)
        {
            pt.x = line->v2->x;
            pt.y = line->v2->y;
            li.p1 = pt;
            pt.x = line->v1->x;
            pt.y = line->v1->y;
            li.p2 = pt;
            li.linedef = i;
            li.side = 1;
            li.offset = 0;
            li.grouped = false;
            segstore->addElement(&li);
        }
    }
}

/*
=====================
=
= BuildBSP
=
=====================
*/

bspnode_t*  startnode;

void SaveDoomMap(void);

void BuildBSP(void)
{
    PrintDialog("BSPc[\z...", hProgDialog, IDC_STEXT1);

    MakeSegs();
    cuts = 0;
    startnode = BSPList(segstore);

    if (startnode->lines == segstore)
        Alert("}bvɂ͓ʌ`̃ZN^[邾Ȃ̂ŁAQ[vO̕ŐȂ\܂");

    Debug("cuts = %d\n", cuts);

    SaveDoomMap();
}
